#include "Pointing.h"

#include <armarx/control/pointing/skill_ids/skill_ids.h>

#include "aron_conversions.h"

namespace armarx::control::pointing::skills
{

    armarx::skills::SkillDescription
    Pointing::GetSkillDescription()
    {
        ParamType defaultParams;
        defaultParams.side = core::Side::RIGHT;
        defaultParams.frame = "root";
        defaultParams.handShape = std::make_optional("Pointing");
        defaultParams.maxVelocity = 0.4;
        defaultParams.maxAcceleration = 0.3;

        return armarx::skills::SkillDescription{.skillId = skill_ids::Pointing,
                                                .description = "Lets the robot point at something.",
                                                .rootProfileDefaults = defaultParams.toAron(),
                                                .timeout = armarx::core::time::Duration::Minutes(1),
                                                .parametersType = ParamType::ToAronType()};
    }

    Pointing::Pointing(const Remote& remote) : Base(GetSkillDescription()), remote_(remote)
    {
    }

    Pointing::MainResult
    Pointing::main(const SpecializedMainInput& in)
    {
        core::Pointing::Parameters parameters = fromAron(in.parameters);
        impl_.emplace(remote_);
        try
        {
            impl_->execute(parameters);
        }
        catch (core::Pointing::PointingFailedException&)
        {
            return MakeFailedResult();
        }
        catch (core::Pointing::PointingAbortedException&)
        {
            return MakeAbortedResult();
        }
        return MakeSucceededResult();
    }

    void
    Pointing::onStopRequested()
    {
        impl_->abort();
    }

    void
    Pointing::onTimeoutReached()
    {
        impl_->abort();
    }

} // namespace armarx::control::pointing::skills
